@@ -239,7 +239,9 @@ class QueryBuilderEnvironment:
         execute_after TIMESTAMP WITH TIME ZONE DEFAULT NOW() NOT NULL,
         status {self.settings.queue_status_type} NOT NULL,
         entrypoint TEXT NOT NULL,
-        payload BYTEA
+        dedupe_key TEXT,
+        payload BYTEA,
+        headers JSONB
     );
     CREATE INDEX {self.settings.queue_table}_priority_id_id1_idx ON {self.settings.queue_table} (priority ASC, id DESC)
         INCLUDE (id) WHERE status = 'queued';
@@ -247,6 +249,8 @@ class QueryBuilderEnvironment:
         INCLUDE (id) WHERE status = 'picked';
     CREATE INDEX {self.settings.queue_table}_queue_manager_id_idx ON {self.settings.queue_table} (queue_manager_id)
         WHERE queue_manager_id IS NOT NULL;
+    CREATE UNIQUE INDEX IF NOT EXISTS {self.settings.queue_table}_unique_dedupe_key ON
+        {self.settings.queue_table} (dedupe_key) WHERE ((status IN ('queued', 'picked') AND dedupe_key IS NOT NULL));
 
     CREATE {durability_policy.queue_log_table} TABLE {self.settings.queue_table_log} (
         id BIGINT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
@@ -255,11 +259,13 @@ class QueryBuilderEnvironment:
         status {self.settings.queue_status_type} NOT NULL,
         priority INT NOT NULL,
         entrypoint TEXT NOT NULL,
+        traceback JSONB DEFAULT NULL,
         aggregated BOOLEAN DEFAULT FALSE
     );
     CREATE INDEX {self.settings.queue_table_log}_not_aggregated ON {self.settings.queue_table_log} ((1)) WHERE not aggregated;
     CREATE INDEX {self.settings.queue_table_log}_created ON {self.settings.queue_table_log} (created);
     CREATE INDEX {self.settings.queue_table_log}_status ON {self.settings.queue_table_log} (status);
+    CREATE INDEX {self.settings.queue_table_log}_job_id_status ON {self.settings.queue_table_log} (job_id, created DESC);
 
     CREATE {durability_policy.statistics_table} TABLE {self.settings.statistics_table} (
         id SERIAL PRIMARY KEY,
@@ -440,9 +446,14 @@ class QueryBuilderEnvironment:
         entrypoint TEXT NOT NULL,
         aggregated BOOLEAN DEFAULT FALSE
     );"""
-        yield f"""CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_not_aggregated ON {self.settings.queue_table_log} ((1)) WHERE not aggregated;"""  # noqa
-        yield f"""CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_created ON {self.settings.queue_table_log} (created);"""  # noqa
-        yield f"""CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_status ON {self.settings.queue_table_log} (status);"""  # noqa
+        yield f"CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_not_aggregated ON {self.settings.queue_table_log} ((1)) WHERE not aggregated;"  # noqa
+        yield f"CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_created ON {self.settings.queue_table_log} (created);"  # noqa
+        yield f"CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_status ON {self.settings.queue_table_log} (status);"  # noqa
+        yield f"ALTER TABLE {self.settings.queue_table_log} ADD COLUMN IF NOT EXISTS traceback JSONB DEFAULT NULL;"  # noqa: E501
+        yield f"ALTER TABLE {self.settings.queue_table} ADD COLUMN IF NOT EXISTS dedupe_key TEXT DEFAULT NULL;"  # noqa: E501
+        yield f"CREATE UNIQUE INDEX IF NOT EXISTS {self.settings.queue_table}_unique_dedupe_key ON {self.settings.queue_table} (dedupe_key) WHERE ((status IN ('queued', 'picked') AND dedupe_key IS NOT NULL));"  # noqa
+        yield f"CREATE INDEX IF NOT EXISTS {self.settings.queue_table_log}_job_id_status ON {self.settings.queue_table_log} (job_id, created DESC);"  # noqa: E501
+        yield f"ALTER TABLE {self.settings.queue_table} ADD COLUMN IF NOT EXISTS headers JSONB;"  # noqa: E501
 
     def build_table_has_column_query(self) -> str:
         """
@@ -459,6 +470,23 @@ class QueryBuilderEnvironment:
             AND column_name = $2
         );"""
 
+    def build_table_has_index_query(self) -> str:
+        """
+        A query to check if a specific index exists in a table.
+
+        Returns:
+            str: The SQL query string to check for a index's existence.
+        """
+        return """SELECT EXISTS (
+            SELECT 1
+            FROM pg_indexes
+            WHERE
+                    tablename  = $1
+                AND indexname  = $2
+                AND schemaname = current_schema()
+        );
+        """
+
     def build_has_table_query(self) -> str:
         """
         A query to check if a specific table exists in a table.
@@ -473,6 +501,27 @@ class QueryBuilderEnvironment:
             AND table_name = $1
         );"""
 
+    def build_has_function_query(self) -> str:
+        """SQL to check if a function exists in the current schema."""
+        return """SELECT EXISTS (
+            SELECT 1
+            FROM pg_proc p
+            JOIN pg_namespace n ON p.pronamespace = n.oid
+            WHERE n.nspname = current_schema()
+              AND p.proname = $1
+        );"""
+
+    def build_has_trigger_query(self) -> str:
+        """SQL to check if a trigger exists in the current schema."""
+        return """SELECT EXISTS (
+            SELECT 1
+            FROM pg_trigger t
+            JOIN pg_class c ON t.tgrelid = c.oid
+            JOIN pg_namespace n ON c.relnamespace = n.oid
+            WHERE n.nspname = current_schema()
+              AND t.tgname = $1
+        );"""
+
     def build_user_types_query(self) -> str:
         """
         A query to list user-defined ENUM types and their labels.
@@ -495,6 +544,100 @@ class QueryBuilderEnvironment:
         yield f"""ALTER TABLE {self.settings.statistics_table} SET {"LOGGED" if durability.statistics_table == "" else "UNLOGGED"};"""  # noqa
         yield f"""ALTER TABLE {self.settings.schedules_table} SET {"LOGGED" if durability.schedules_table == "" else "UNLOGGED"};"""  # noqa
 
+    def build_optimize_autovacuum_query(self) -> str:
+        """SQL for recommended autovacuum settings."""
+
+        return f"""ALTER TABLE {self.settings.queue_table} SET (
+    /* vacuum very early: after ~1 %% or 1 000 dead rows       */
+    autovacuum_vacuum_scale_factor   = 0.01,
+    autovacuum_vacuum_threshold      = 1000,
+
+    /* analyse even earlier so the planner keeps up-to-date   */
+    autovacuum_analyze_scale_factor  = 0.02,
+    autovacuum_analyze_threshold     = 500,
+
+    /* work fast: spend ≈10× the default IO budget            */
+    autovacuum_vacuum_cost_limit     = 10000,
+    autovacuum_vacuum_cost_delay     = 0,      -- no 20 ms naps
+
+    /* leave headroom for HOT updates so fewer dead tuples    */
+    fillfactor                       = 70
+);
+
+ALTER TABLE {self.settings.schedules_table} SET (
+/* vacuum very early: after ~1 %% or 1 000 dead rows       */
+autovacuum_vacuum_scale_factor   = 0.01,
+autovacuum_vacuum_threshold      = 1000,
+
+/* analyse even earlier so the planner keeps up-to-date   */
+autovacuum_analyze_scale_factor  = 0.02,
+autovacuum_analyze_threshold     = 500,
+
+/* work fast: spend ≈10× the default IO budget            */
+autovacuum_vacuum_cost_limit     = 10000,
+autovacuum_vacuum_cost_delay     = 0,      -- no 20 ms naps
+
+/* leave headroom for HOT updates so fewer dead tuples    */
+fillfactor                       = 70
+);
+
+ALTER TABLE {self.settings.queue_table_log} SET (
+    /* essentially disable vacuum (but keep freeze safety)    */
+    autovacuum_vacuum_scale_factor   = 0.95,
+    autovacuum_vacuum_threshold      = 1000000,
+
+    /* still analyse at ~5 %% growth so estimates stay OK      */
+    autovacuum_analyze_scale_factor  = 0.05,
+    autovacuum_analyze_threshold     = 10000
+);
+
+ALTER TABLE {self.settings.statistics_table} SET (
+    /* essentially disable vacuum (but keep freeze safety)    */
+    autovacuum_vacuum_scale_factor   = 0.95,
+    autovacuum_vacuum_threshold      = 1000000,
+
+    /* still analyse at ~5 %% growth so estimates stay OK      */
+    autovacuum_analyze_scale_factor  = 0.05,
+    autovacuum_analyze_threshold     = 10000
+);"""
+
+    def build_optimize_autovacuum_rollback_query(self) -> str:
+        """SQL to reset autovacuum settings."""
+
+        return f"""ALTER TABLE {self.settings.queue_table} RESET (
+    autovacuum_vacuum_scale_factor,
+    autovacuum_vacuum_threshold,
+    autovacuum_analyze_scale_factor,
+    autovacuum_analyze_threshold,
+    autovacuum_vacuum_cost_limit,
+    autovacuum_vacuum_cost_delay,
+    fillfactor
+);
+
+ALTER TABLE {self.settings.schedules_table} RESET (
+autovacuum_vacuum_scale_factor,
+autovacuum_vacuum_threshold,
+autovacuum_analyze_scale_factor,
+autovacuum_analyze_threshold,
+autovacuum_vacuum_cost_limit,
+autovacuum_vacuum_cost_delay,
+fillfactor
+);
+
+ALTER TABLE {self.settings.queue_table_log} RESET (
+    autovacuum_vacuum_scale_factor,
+    autovacuum_vacuum_threshold,
+    autovacuum_analyze_scale_factor,
+    autovacuum_analyze_threshold
+);
+
+ALTER TABLE {self.settings.statistics_table} RESET (
+    autovacuum_vacuum_scale_factor,
+    autovacuum_vacuum_threshold,
+    autovacuum_analyze_scale_factor,
+    autovacuum_analyze_threshold
+);"""
+
 
 @dataclasses.dataclass
 class QueryQueueBuilder:
@@ -544,6 +687,15 @@ class QueryQueueBuilder:
             AND entrypoint = ANY($2)
         GROUP BY entrypoint
     ),
+    jobs_by_queue_manager AS (
+        SELECT
+            COUNT(*) AS qm_count
+        FROM {self.settings.queue_table}
+        WHERE
+                queue_manager_id IS NOT NULL
+            AND queue_manager_id = $6
+            AND entrypoint = ANY($2)
+    ),
     next_job_queued AS (
         SELECT {self.settings.queue_table}.id
         FROM {self.settings.queue_table}
@@ -555,9 +707,9 @@ class QueryQueueBuilder:
                 {self.settings.queue_table}.entrypoint = ANY($2)
             AND {self.settings.queue_table}.status = 'queued'
             AND {self.settings.queue_table}.execute_after < NOW()
-            AND (
-                NOT entrypoint_execution_params.serialized
-                OR NOT EXISTS (
+            AND ($7::BIGINT IS NULL OR (SELECT qm_count FROM jobs_by_queue_manager) < $7)
+            AND NOT (
+                entrypoint_execution_params.serialized AND EXISTS (
                     SELECT 1
                     FROM {self.settings.queue_table} existing_job
                     WHERE existing_job.entrypoint = {self.settings.queue_table}.entrypoint
@@ -584,6 +736,25 @@ class QueryQueueBuilder:
             AND {self.settings.queue_table}.execute_after < NOW()
             AND entrypoint_execution_params.retry_after > interval '0'
             AND {self.settings.queue_table}.heartbeat < NOW() - entrypoint_execution_params.retry_after
+            AND ($7::BIGINT IS NULL OR (SELECT qm_count FROM jobs_by_queue_manager) < $7)
+            -- Honor serialized_dispatch on retries: skip retry if serialized and a job is still in flight
+            AND NOT (
+                entrypoint_execution_params.serialized AND EXISTS (
+                    SELECT 1
+                    FROM {self.settings.queue_table} existing_job
+                    WHERE existing_job.entrypoint = {self.settings.queue_table}.entrypoint
+                    AND existing_job.status = 'picked'
+                )
+            )
+            -- Also enforce concurrency_limit on retry
+            AND (
+                entrypoint_execution_params.concurrency_limit <= 0
+                OR EXISTS(
+                    SELECT 1 FROM jobs_by_queue_manager_entrypoint j
+                    WHERE j.entrypoint = {self.settings.queue_table}.entrypoint
+                      AND j.count < entrypoint_execution_params.concurrency_limit
+                )
+            )
         ORDER BY {self.settings.queue_table}.heartbeat DESC, {self.settings.queue_table}.id ASC
         FOR UPDATE SKIP LOCKED
         LIMIT $1
@@ -612,6 +783,17 @@ class QueryQueueBuilder:
     SELECT * FROM updated ORDER BY priority DESC, id ASC;
     """  # noqa
 
+    def build_has_queued_work(self) -> str:
+        return f"""
+        SELECT
+            COUNT(*) AS queued_work
+        FROM
+            {self.settings.queue_table}
+        WHERE
+                entrypoint = ANY($1)
+            AND status = 'queued'
+        """
+
     def build_enqueue_query(self) -> str:
         """
         Generate SQL query to insert new jobs into the queue.
@@ -626,12 +808,14 @@ class QueryQueueBuilder:
         return f"""
         WITH inserted AS (
             INSERT INTO {self.settings.queue_table}
-            (priority, entrypoint, payload, execute_after, status)
+            (priority, entrypoint, payload, execute_after, dedupe_key, headers, status)
             VALUES (
                 UNNEST($1::int[]),                  -- priority
                 UNNEST($2::text[]),                 -- entrypoint
                 UNNEST($3::bytea[]),                -- payload
                 UNNEST($4::interval[]) + NOW(),     -- execute_after
+                UNNEST($5::text[]),                 -- dedupe_key
+                UNNEST($6::jsonb[]),                -- headers
                 'queued'                            -- status
             )
             RETURNING id, entrypoint, status, priority
@@ -721,24 +905,27 @@ class QueryQueueBuilder:
         ), job_status AS (
             SELECT
                 UNNEST($1::integer[])                           AS id,
-                UNNEST($2::{self.settings.queue_status_type}[]) AS status
+                UNNEST($2::{self.settings.queue_status_type}[]) AS status,
+                UNNEST($3::JSONB[])                             AS traceback
         ), merged AS (
             SELECT
-                job_status.id       AS id,
-                job_status.status   AS status,
-                deleted.entrypoint  AS entrypoint,
-                deleted.priority    AS priority
+                job_status.id           AS id,
+                job_status.status       AS status,
+                job_status.traceback    AS traceback,
+                deleted.entrypoint      AS entrypoint,
+                deleted.priority        AS priority
             FROM job_status
             INNER JOIN deleted
-            ON deleted.id = job_status.id
+                ON deleted.id = job_status.id
         )
         INSERT INTO {self.settings.queue_table_log} (
             job_id,
             status,
             entrypoint,
-            priority
+            priority,
+            traceback
         )
-        SELECT id, status, entrypoint, priority FROM merged
+        SELECT id, status, entrypoint, priority, traceback FROM merged
         """
 
     def build_truncate_log_statistics_query(self) -> str:
@@ -858,6 +1045,13 @@ class QueryQueueBuilder:
             count = {self.settings.statistics_table}.count + EXCLUDED.count
         """  # noqa
 
+    def build_job_status_query(self) -> str:
+        return f"""SELECT DISTINCT ON (job_id) job_id, status
+        FROM   {self.settings.queue_table_log}
+        WHERE  job_id = ANY($1)
+        ORDER  BY job_id, created DESC, id DESC;
+        """
+
 
 @dataclasses.dataclass
 class QuerySchedulerBuilder:
