@@ -11,12 +11,10 @@ from __future__ import annotations
 import asyncio
 import contextlib
 import dataclasses
-import functools
 import sys
 import uuid
-import warnings
 from collections import Counter, deque
-from contextlib import nullcontext
+from contextlib import nullcontext, suppress
 from datetime import timedelta
 from math import isfinite
 from typing import AsyncGenerator, Callable
@@ -25,7 +23,9 @@ import anyio
 
 from . import (
     buffers,
+    cache,
     db,
+    errors,
     executors,
     heartbeat,
     helpers,
@@ -35,11 +35,10 @@ from . import (
     qb,
     queries,
     tm,
+    tracing,
     types,
 )
 
-warnings.simplefilter("default", DeprecationWarning)
-
 
 @dataclasses.dataclass
 class QueueManager:
@@ -60,6 +59,9 @@ class QueueManager:
         queue_manager_id (uuid.UUID): Unique identifier for each QueueManager instance.
         job_context (dict[models.JobId, models.Context]): Contexts for jobs,
             including cancellation scopes.
+        shutdown_on_listener_failure
+            If *True*, raise `FailingListenerError` and set the shutdown event when
+            the periodic listener health‑check times out.
     """
 
     connection: db.Driver
@@ -93,16 +95,64 @@ class QueueManager:
         default_factory=dict,
     )
 
-    @property
-    def alive(self) -> asyncio.Event:
-        # For backwards compatibility
-        warnings.warn(
-            "The `alive` property is deprecated and will be removed in a future release. "
-            "Please use `shutdown` instead.",
-            DeprecationWarning,
-            stacklevel=2,
+    pending_health_check: dict[uuid.UUID, asyncio.Future[models.HealthCheckEvent]] = (
+        dataclasses.field(
+            init=False,
+            default_factory=dict,
         )
-        return self.shutdown
+    )
+
+    async def listener_healthy(
+        self,
+        timeout: timedelta = timedelta(seconds=10),
+    ) -> models.HealthCheckEvent:
+        """
+        Perform a health check by sending a notification and waiting for a response.
+
+        Args:
+            timeout (timedelta): Maximum time to wait for the health check response.
+
+        Returns:
+            models.HealthCheckEvent: The received health check event.
+        """
+        health_check_event_id = uuid.uuid4()
+        fut = asyncio.Future[models.HealthCheckEvent]()
+        self.pending_health_check[health_check_event_id] = fut
+
+        await self.queries.notify_health_check(health_check_event_id)
+        try:
+            return await asyncio.wait_for(
+                asyncio.shield(fut),
+                timeout.total_seconds(),
+            )
+        except (TimeoutError, asyncio.TimeoutError):
+            fut.set_exception(errors.FailingListenerError)
+            return await fut
+        except Exception as e:
+            fut.set_exception(e)
+            return await fut
+        finally:
+            self.pending_health_check.pop(health_check_event_id, None)
+
+    async def _run_periodic_health_check(
+        self,
+        interval: timedelta = timedelta(seconds=10),
+    ) -> None:
+        """
+        Periodically perform a health check by calling `listener_healthy`.
+
+        Args:
+            interval (timedelta): Time interval between health checks.
+            shutdown_on_failure (bool): Whether to set the shutdown event if a health check fails.
+        """
+
+        while not self.shutdown.is_set():
+            await self.listener_healthy(timeout=interval)
+            with suppress(TimeoutError, asyncio.TimeoutError):
+                await asyncio.wait_for(
+                    self.shutdown.wait(),
+                    timeout=interval.total_seconds(),
+                )
 
     def __post_init__(self) -> None:
         """
@@ -158,7 +208,6 @@ class QueueManager:
         concurrency_limit: int = 0,
         retry_timer: timedelta = timedelta(seconds=0),
         serialized_dispatch: bool = False,
-        executor: type[executors.AbstractEntrypointExecutor] | None = None,
         executor_factory: Callable[
             [executors.EntrypointExecutorParameters],
             executors.AbstractEntrypointExecutor,
@@ -177,7 +226,6 @@ class QueueManager:
             concurrency_limit (int): Max number of concurrent jobs allowed for this entrypoint.
             retry_timer (timedelta): Duration to wait before retrying 'picked' jobs.
             serialized_dispatch (bool): Whether to serialize dispatching of jobs.
-            executor (JobExecutor): Custom executor instance to use.
 
         Returns:
             Callable[[T], T]: A decorator that registers the function as an entrypoint.
@@ -187,14 +235,6 @@ class QueueManager:
             ValueError: If `requests_per_second` or `concurrency_limit` are negative.
         """
 
-        if executor is not None:
-            warnings.warn(
-                "The 'executor' parameter is deprecated and will be removed in a future version. "
-                "Please use 'executor_factory' instead for custom executor handling.",
-                DeprecationWarning,
-                stacklevel=3,
-            )
-
         if name in self.entrypoint_registry:
             raise RuntimeError(f"{name} already in registry, name must be unique.")
 
@@ -292,7 +332,11 @@ class QueueManager:
             if below_capacity_limit(entrypoint, executor)
         }
 
-    async def fetch_jobs(self, batch_size: int) -> AsyncGenerator[models.Job, None]:
+    async def fetch_jobs(
+        self,
+        batch_size: int,
+        global_concurrency_limit: int | None,
+    ) -> AsyncGenerator[models.Job, None]:
         """
         Fetch jobs from the queue that are ready for processing, yielding them one at a time.
 
@@ -316,6 +360,7 @@ class QueueManager:
                     batch_size=batch_size,
                     entrypoints=entrypoints,
                     queue_manager_id=self.queue_manager_id,
+                    global_concurrency_limit=global_concurrency_limit,
                 )
             ):
                 break
@@ -362,6 +407,8 @@ class QueueManager:
             (self.queries.qbe.settings.queue_table, "heartbeat"),
             (self.queries.qbe.settings.queue_table, "queue_manager_id"),
             (self.queries.qbe.settings.queue_table, "execute_after"),
+            (self.queries.qbe.settings.queue_table, "headers"),
+            (self.queries.qbe.settings.queue_table_log, "traceback"),
         ):
             if not (await self.queries.table_has_column(table, column)):
                 raise RuntimeError(
@@ -375,11 +422,25 @@ class QueueManager:
                     f"The {enum} is missing the '{key}' type, please run 'pgq upgrade'"
                 )
 
+        for table, index in (
+            (
+                self.queries.qbe.settings.queue_table_log,
+                f"{self.queries.qbe.settings.queue_table_log}_job_id_status",
+            ),
+        ):
+            if not (await self.queries.table_has_index(table, index)):
+                raise RuntimeError(
+                    f"The required index '{index}' is missing in the '{table}' table. "
+                    f"Please run 'pgq upgrade' to ensure all schema changes are applied."
+                )
+
     async def run(
         self,
         dequeue_timeout: timedelta = timedelta(seconds=30),
         batch_size: int = 10,
         mode: types.QueueExecutionMode = types.QueueExecutionMode.continuous,
+        max_concurrent_tasks: int | None = None,
+        shutdown_on_listener_failure: bool = False,
     ) -> None:
         """
         Run the main loop to process jobs from the queue.
@@ -390,75 +451,93 @@ class QueueManager:
         Args:
             dequeue_timeout (timedelta): Timeout duration for waiting to dequeue jobs.
             batch_size (int): Number of jobs to retrieve in each batch.
-            burst_mode (bool): Whether to run in burst mode, fetch until
+            mode (QueueExecutionMode): Whether to run in `continuous` or `drain` until
                 queue is empty then shutdown.
-
+            max_concurrent_tasks (int|None, default=None): Limit the total number of tasks that
+                can run at the same time. If unspecified or None, there is no limit.
         Raises:
             RuntimeError: If required database columns or types are missing.
         """
 
         await self.verify_structure()
 
-        job_status_log_buffer_timeout = timedelta(seconds=0.01)
         heartbeat_buffer_timeout = helpers.retry_timer_buffer_timeout(
             [x.parameters.retry_timer for x in self.entrypoint_registry.values()]
         )
 
+        max_concurrent_tasks = max_concurrent_tasks or sys.maxsize
+
+        if max_concurrent_tasks < 2 * batch_size:
+            raise RuntimeError("max_concurrent_tasks must be at least twice the batch size.")
+
         async with (
             buffers.JobStatusLogBuffer(
                 max_size=batch_size,
-                timeout=job_status_log_buffer_timeout,
                 callback=self.queries.log_jobs,
             ) as jbuff,
             buffers.HeartbeatBuffer(
-                max_size=sys.maxsize,
-                timeout=heartbeat_buffer_timeout / 2,
+                # Flush will be mainly driven by timeouts, but allow flush if
+                # backlog becomes too large.
+                max_size=batch_size**2,
+                timeout=heartbeat_buffer_timeout / 4,
                 callback=self.queries.update_heartbeat,
             ) as hbuff,
             buffers.RequestsPerSecondBuffer(
                 max_size=batch_size,
-                timeout=timedelta(seconds=0.01),
                 callback=self.update_rps_stats,
             ) as rpsbuff,
             tm.TaskManager() as task_manager,
             self.connection,
         ):
+            periodic_health_check_task = asyncio.create_task(self._run_periodic_health_check())
+
             notice_event_listener = listeners.PGNoticeEventListener()
             await listeners.initialize_notice_event_listener(
                 self.connection,
                 self.channel,
-                functools.partial(
-                    listeners.handle_event_type,
+                listeners.default_event_router(
                     notice_event_queue=notice_event_listener,
                     statistics=self.entrypoint_statistics,
                     canceled=self.job_context,
+                    pending_health_check=self.pending_health_check,
                 ),
             )
 
             shutdown_task = asyncio.create_task(self.shutdown.wait())
             event_task: None | asyncio.Task[None | models.TableChangedEvent] = None
+            cached_queued_work = cache.TTLCache.create(
+                ttl=timedelta(seconds=0.250),
+                on_expired=lambda: self.queries.queued_work(list(self.entrypoint_registry.keys())),
+            )
 
             while not self.shutdown.is_set():
-                async for job in self.fetch_jobs(batch_size):
+                async for job in self.fetch_jobs(batch_size, max_concurrent_tasks):
                     await rpsbuff.add(job.entrypoint)
                     self.job_context[job.id] = models.Context(cancellation=anyio.CancelScope())
-                    task_manager.add(
-                        asyncio.create_task(
-                            self._dispatch(
-                                job,
-                                jbuff,
-                                hbuff,
-                            )
-                        )
-                    )
+                    task_manager.add(asyncio.create_task(self._dispatch(job, jbuff, hbuff)))
 
                     with contextlib.suppress(asyncio.QueueEmpty):
                         notice_event_listener.get_nowait()
 
-                # Run until the queue is empty and then shutdown.
-                if mode is types.QueueExecutionMode.drain:
+                    if self.shutdown.is_set():
+                        break
+
+                # Run until the queue is empty and then shutdown,
+                # if max_concurrent_tasks is low, we could exit early due to
+                # fetch_jobs not yield more jobs due to at work capacity. Need to check
+                # back in with the database to see if there are any jobs left.
+                if mode is types.QueueExecutionMode.drain and (await cached_queued_work()) == 0:
                     self.shutdown.set()
 
+                if (
+                    periodic_health_check_task.done()
+                    and periodic_health_check_task.exception()
+                    and mode is not types.QueueExecutionMode.drain
+                    and shutdown_on_listener_failure
+                ):
+                    self.shutdown.set()
+                    await periodic_health_check_task
+
                 event_task = helpers.wait_for_notice_event(
                     notice_event_listener,
                     dequeue_timeout,
@@ -500,10 +579,16 @@ class QueueManager:
 
         executor = self.entrypoint_registry[job.entrypoint]
 
+        # Send heartbeats several times within ``retry_timer`` to keep the job
+        # alive while it is running.
+        trace_context = (
+            tracing.TRACER.tracer.trace_process(job) if tracing.TRACER.tracer else nullcontext()
+        )
         async with (
+            trace_context,
             heartbeat.Heartbeat(
                 job.id,
-                executor.parameters.retry_timer,
+                executor.parameters.retry_timer / 2,
                 hbuff,
             ),
             self.entrypoint_statistics[job.entrypoint].concurrency_limiter,
@@ -515,13 +600,21 @@ class QueueManager:
                 ctx = self.get_context(job.id)
                 if not ctx.cancellation.cancel_called:
                     await executor.execute(job, ctx)
-            except Exception:
+            except Exception as e:
                 logconfig.logger.exception(
                     "Exception while processing entrypoint/job-id: %s/%s",
                     job.entrypoint,
                     job.id,
                 )
-                await jbuff.add((job, "exception"))
+                tbr = models.TracebackRecord.from_exception(
+                    exc=e,
+                    job_id=job.id,
+                    additional_context={
+                        "entrypoint": job.entrypoint,
+                        "queue_manager_id": self.queue_manager_id,
+                    },
+                )
+                await jbuff.add((job, "exception", tbr))
             else:
                 logconfig.logger.debug(
                     "Dispatching entrypoint/id: %s/%s - successful",
@@ -529,6 +622,6 @@ class QueueManager:
                     job.id,
                 )
                 canceled = ctx.cancellation.cancel_called
-                await jbuff.add((job, "canceled" if canceled else "successful"))
+                await jbuff.add((job, "canceled" if canceled else "successful", None))
             finally:
                 self.job_context.pop(job.id, None)
