@@ -12,9 +12,14 @@ class QueueExecutionMode(Enum):
 
 ###### Events ######
 Channel = NewType("Channel", str)
-PGChannel = Channel  # TODO: Depricate
+PGChannel = Channel  # TODO: Deprecate
 OPERATIONS = Literal["insert", "update", "delete", "truncate"]
-EVENT_TYPES = Literal["table_changed_event", "requests_per_second_event", "cancellation_event"]
+EVENT_TYPES = Literal[
+    "table_changed_event",
+    "requests_per_second_event",
+    "cancellation_event",
+    "health_check_event",
+]
 
 
 ###### Jobs ######
