@@ -18,6 +18,7 @@ def mocked_job(
     entrypoint: str = "test",
     payload: bytes | None = None,
     queue_manager_id: None | uuid.UUID = None,
+    headers: dict | None = None,
 ) -> Job:
     now = datetime.now(timezone.utc)
     return Job(
@@ -31,4 +32,5 @@ def mocked_job(
         entrypoint=entrypoint,
         payload=payload,
         queue_manager_id=queue_manager_id or uuid.uuid4(),
+        headers=headers,
     )
