@@ -2,9 +2,10 @@ import asyncio
 import uuid
 from datetime import timedelta
 
+import asyncpg
 import pytest
 
-from pgqueuer import db, models, queries
+from pgqueuer import db, errors, models, queries
 
 
 @pytest.mark.parametrize("N", (1, 2, 64))
@@ -39,12 +40,13 @@ async def test_queries_next_jobs(
             "placeholder": queries.EntrypointExecutionParameter(timedelta(days=1), False, 0)
         },
         queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
     ):
         for job in jobs:
             payoad = job.payload
             assert payoad is not None
             seen.append(int(payoad))
-            await q.log_jobs([(job, "successful")])
+            await q.log_jobs([(job, "successful", None)])
 
     assert seen == list(range(N))
 
@@ -73,12 +75,13 @@ async def test_queries_next_jobs_concurrent(
             },
             batch_size=10,
             queue_manager_id=uuid.uuid4(),
+            global_concurrency_limit=1000,
         ):
             for job in jobs:
                 payload = job.payload
                 assert payload is not None
                 seen.append(int(payload))
-                await q.log_jobs([(job, "successful")])
+                await q.log_jobs([(job, "successful", None)])
 
     await asyncio.wait_for(
         asyncio.gather(*[consumer() for _ in range(concurrency)]),
@@ -119,9 +122,10 @@ async def test_move_job_log(
             "placeholder": queries.EntrypointExecutionParameter(timedelta(days=1), False, 0)
         },
         queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
     ):
         for job in jobs:
-            await q.log_jobs([(job, "successful")])
+            await q.log_jobs([(job, "successful", None)])
 
     assert sum(x.status == "successful" for x in await q.queue_log()) == N
 
@@ -194,10 +198,11 @@ async def test_queue_priority(
         },
         batch_size=10,
         queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
     ):
         for job in next_jobs:
             jobs.append(job)
-            await q.log_jobs([(job, "successful")])
+            await q.log_jobs([(job, "successful", None)])
 
     assert jobs == sorted(jobs, key=lambda x: x.priority, reverse=True)
 
@@ -224,6 +229,7 @@ async def test_queue_retry_timer(
             "placeholder": queries.EntrypointExecutionParameter(timedelta(days=1), False, 0)
         },
         queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
     ):
         ...
 
@@ -235,6 +241,7 @@ async def test_queue_retry_timer(
                     "placeholder": queries.EntrypointExecutionParameter(timedelta(days=1), False, 0)
                 },
                 queue_manager_id=uuid.uuid4(),
+                global_concurrency_limit=1000,
             ),
         )
         == 0
@@ -248,6 +255,7 @@ async def test_queue_retry_timer(
         entrypoints={"placeholder": queries.EntrypointExecutionParameter(retry_timer, False, 0)},
         batch_size=10,
         queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
     ):
         jobs.extend(next_jobs)
 
@@ -280,13 +288,14 @@ async def test_queue_log_queued_picked_successful(
             )
         },
         queue_manager_id=queue_manager_id,
+        global_concurrency_limit=1000,
     ):
         picked_jobs.extend(jobs)
 
     assert sum(x.status == "picked" for x in await q.queue_log()) == N
 
     for job in picked_jobs:
-        await q.log_jobs([(job, "successful")])
+        await q.log_jobs([(job, "successful", None)])
 
     assert sum(x.status == "successful" for x in await q.queue_log()) == N
 
@@ -317,12 +326,142 @@ async def test_queue_log_queued_picked_exception(
             )
         },
         queue_manager_id=queue_manager_id,
+        global_concurrency_limit=1000,
     ):
         picked_jobs.extend(jobs)
 
     assert sum(x.status == "picked" for x in await q.queue_log()) == N
 
     for job in picked_jobs:
-        await q.log_jobs([(job, "exception")])
+        await q.log_jobs([(job, "exception", None)])
 
     assert sum(x.status == "exception" for x in await q.queue_log()) == N
+
+
+async def test_queue_log_queued_dedupe_key_raises(
+    apgdriver: db.Driver,
+    pgdriver: db.SyncDriver,
+) -> None:
+    aq = queries.Queries(apgdriver)
+
+    await aq.enqueue(
+        "test_queue_log_queued_dedupe_key_raises",
+        None,
+        dedupe_key="test_queue_log_queued_dedupe_key_raises",
+    )
+    assert sum(x.count for x in await aq.queue_size()) == 1
+
+    with pytest.raises(errors.DuplicateJobError):
+        await aq.enqueue(
+            "test_queue_log_queued_dedupe_key_raises",
+            None,
+            dedupe_key="test_queue_log_queued_dedupe_key_raises",
+        )
+    assert sum(x.count for x in await aq.queue_size()) == 1
+
+    sq = queries.SyncQueries(pgdriver)
+    with pytest.raises(errors.DuplicateJobError):
+        sq.enqueue(
+            "test_queue_log_queued_dedupe_key_raises",
+            None,
+            dedupe_key="test_queue_log_queued_dedupe_key_raises",
+        )
+
+    assert sum(x.count for x in sq.queue_size()) == 1
+
+
+async def test_queue_log_queued_dedupe_key_raises_array(
+    apgdriver: db.Driver,
+    pgdriver: db.SyncDriver,
+) -> None:
+    aq = queries.Queries(apgdriver)
+
+    await aq.enqueue(
+        ["test_queue_log_queued_dedupe_key_raises_array"],
+        [None],
+        [0],
+        dedupe_key=["test_queue_log_queued_dedupe_key_raises_array"],
+    )
+    assert sum(x.count for x in await aq.queue_size()) == 1
+
+    with pytest.raises(errors.DuplicateJobError):
+        await aq.enqueue(
+            ["test_queue_log_queued_dedupe_key_raises_array"],
+            [None],
+            [0],
+            dedupe_key=["test_queue_log_queued_dedupe_key_raises_array"],
+        )
+
+    assert sum(x.count for x in await aq.queue_size()) == 1
+
+    sq = queries.SyncQueries(pgdriver)
+    with pytest.raises(errors.DuplicateJobError):
+        sq.enqueue(
+            ["test_queue_log_queued_dedupe_key_raises_array"],
+            [None],
+            [0],
+            dedupe_key=["test_queue_log_queued_dedupe_key_raises_array"],
+        )
+
+    assert sum(x.count for x in sq.queue_size()) == 1
+
+
+@pytest.mark.parametrize("batch_size", (2, 10))
+@pytest.mark.parametrize("loop_size", (2, 10))
+async def test_queue_log_queued_dedupe_key_raises_concurrent(
+    batch_size: int,
+    loop_size: int,
+    apgdriver: db.Driver,
+) -> None:
+    async with asyncpg.create_pool() as pool:
+        for _ in range(loop_size):
+            with pytest.raises(errors.DuplicateJobError):
+                await asyncio.gather(
+                    *[
+                        queries.Queries(db.AsyncpgPoolDriver(pool)).enqueue(
+                            "test_queue_log_queued_dedupe_key_raises_concurrent",
+                            None,
+                            dedupe_key="test_queue_log_queued_dedupe_key_raises_concurrent",
+                        )
+                        for _ in range(batch_size)
+                    ],
+                )
+            assert sum(x.count for x in await queries.Queries(apgdriver).queue_size()) == 1
+
+    assert sum(x.count for x in await queries.Queries(apgdriver).queue_size()) == 1
+
+
+async def test_queue_log_queued_dedupe_key_raises_contains_dedupe_key(
+    apgdriver: db.Driver,
+    pgdriver: db.SyncDriver,
+) -> None:
+    dedupe_key = "test_queue_log_queued_dedupe_key_raises_contains_dedupe_key"
+    await queries.Queries(apgdriver).enqueue("...", None, dedupe_key=dedupe_key)
+    assert sum(x.count for x in await queries.Queries(apgdriver).queue_size()) == 1
+
+    with pytest.raises(errors.DuplicateJobError) as raised:
+        await queries.Queries(apgdriver).enqueue("...", None, dedupe_key=dedupe_key)
+    assert dedupe_key in raised.value.dedupe_key
+
+    with pytest.raises(errors.DuplicateJobError) as raised:
+        queries.SyncQueries(pgdriver).enqueue("...", None, dedupe_key=dedupe_key)
+    assert dedupe_key in raised.value.dedupe_key
+
+
+async def test_enqueue_with_headers(apgdriver: db.Driver) -> None:
+    q = queries.Queries(apgdriver)
+    headers = {"trace": "abc"}
+    await q.enqueue("header_task", None, headers=headers)
+
+    jobs = await q.dequeue(
+        entrypoints={
+            "header_task": queries.EntrypointExecutionParameter(timedelta(days=1), False, 0)
+        },
+        batch_size=1,
+        queue_manager_id=uuid.uuid4(),
+        global_concurrency_limit=1000,
+    )
+
+    assert len(jobs) == 1
+    assert jobs[0].headers == headers
+    await q.log_jobs([(jobs[0], "successful", None)])
